import enum
import os
from collections import defaultdict
from dataclasses import dataclass

import yaml

from architest.core import (find_dependencies, find_docstrings,
                            find_extensions, find_instantiations)


@dataclass
class Connection:
    uml: str
    name: str
    lineno: int = 1

    def __hash__(self) -> int:
        return hash(self.name)


class ConnectionType(enum.Enum):
    DEPENDENCY = Connection(uml='..>', name='dependency')
    AGGREGATION = Connection(uml='--o', name='aggregation')
    ASSOCIATION = Connection(uml='-->', name='association')
    COMPOSITION = Connection(uml='--*', name='composition')
    IMPLEMENTATION = Connection(uml='..|>', name='implementation')
    EXTENSION = Connection(uml='--|>', name='extension')


mapping = {
    'dependency': ConnectionType.DEPENDENCY.value,
    'aggregation': ConnectionType.AGGREGATION.value,
    'assoctiation': ConnectionType.ASSOCIATION.value,
    'composition': ConnectionType.COMPOSITION.value,
    'implementation': ConnectionType.IMPLEMENTATION.value,
    'extension': ConnectionType.EXTENSION.value
}


def convert_connection_name_to_class(name: str) -> Connection:
    return mapping[name]


@dataclass
class Path:
    path: str
    lineno: int = 1

    @property
    def upper_module_name(self):
        if '.' in self.path:
            return self.path.split('.')[0]
        return self.path

    # @property
    # def path_split(self):
    # 	return [module.replace('.py', '') for module in self.path.split(os.sep) if module != '..']

    # @property
    # def python_path(self):
    # 	return '.'.join(self.path_split)

    @property
    def python_name(self):
        return self.path.split('.')[-1]

    @property
    def module_path(self):
        if '.' in self.path:
            return self.path[:self.path.rfind('.')]
        return self.path

    def __repr__(self) -> str:
        return repr(f'{self.path}:{self.lineno}')

    def __hash__(self) -> int:
        return hash(self.path)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.path == other.path
        else:
            return NotImplemented


class Module:
    def __init__(self, file_path, module_path) -> None:
        self.path: Path = Path(module_path)
        self.name = self.path.python_name
        self.file_path = file_path
        self.connections = defaultdict(set)
        self._imports = {}
        self._local_paths = {}

    @property
    def id(self):
        return self.path + self.name

    def full_path(self):
        return f'{self.path}.{self.name}'

    def update_imported_module_info(self, path: Path, lineno: int):
        return Path(path.path, lineno)

    @property
    def docstrings(self):
        return find_docstrings(self.file_path)

    @property
    def imports(self):
        if len(self._imports) == 0:
            self._imports = {Path(*path_lineno)
                             for path_lineno in find_dependencies(self.file_path)}
            self._local_paths = {path.python_name: path for path in self._imports}
        return self._imports

    @property
    def local_paths(self) -> dict[str, Path]:
        if not self._local_paths:
            self.imports
        return self._local_paths

    @property
    def dependencies(self):
        return self.imports.difference(self.aggregations)

    @property
    def aggregations(self) -> set[Path]:
        return {self.update_imported_module_info(self.local_paths[name], lineno)
                for name, lineno in find_instantiations(self.file_path) if name in self.local_paths}\
            .intersection(self.imports)

    @property
    def extensions(self):
        return {self.update_imported_module_info(self.local_paths[name], lineno)
                for name, lineno in find_extensions(self.file_path) if name in self.local_paths}\
            .intersection(self.imports)\
            .difference(self.aggregations)

    def __hash__(self) -> int:
        return hash(self.path.path)


@dataclass
class Rule:
    module_from: Module
    connection: Connection
    module_to: Module

    def _dependency_direction(self):
        return f'{self.module_from.path.path} {self.connection.uml} {self.module_to.path.path}'

    def _dependency_path(self):
        return f'{self.module_from.file_path}:{self.connection.lineno}'

    def violation_error(self):
        return f'Illegal {self.connection.name}\t{self._dependency_direction()} @ {self._dependency_path()}'

    def __hash__(self) -> int:
        return hash(self.module_from.path.path + self.connection.uml + self.module_to.path.path)

    def __eq__(self, other) -> bool:
        return other.module_from.path.path == self.module_from.path.path\
            and self.module_to.path.path == other.module_to.path.path\
            and self.connection.uml == other.connection.uml


class RuleSet:
    rules: set[Rule] = set()

    def add(self, module_from: Module, module_to: Module, connection: Connection):
        self.rules.add(Rule(module_from=module_from, module_to=module_to, connection=connection))

    def drop(self, rule: Rule):
        self.rules.remove(rule)


class Project:
    def __init__(self, root: str) -> None:
        self._modules = {}
        self.root: str = root
        self._code_root = ''
        self.is_package = False
        self.raw_rules = None
        self.real_rules = RuleSet()
        self.ignore = []
        self.violations = set()
        self.hidden_connections = []

    def read_config(self):
        file_path = self.root + '/' + '.architest.yml'
        if not os.path.exists(file_path):
            return
        with open(self.root + '/' + '.architest.yml', 'r') as config:
            try:
                self._read_config(config)
            except yaml.YAMLError as exc:
                print(exc)  # noqa: T001

    def _read_config(self, config):
        raw_rules = yaml.safe_load(config)
        self.ignore = raw_rules['ignore']
        self.hidden_connections = self.hidden_connections + raw_rules.get('hide_connections', [])
        self._code_root = raw_rules['code_root']
        self.is_package = raw_rules.get('is_package', False)
        self._code_root = '' if self._code_root is None else self._code_root
        self.raw_rules = RuleSet()

        self.raw_rules = raw_rules['modules']

    def _module_not_in_rules(self, module):
        return module not in \
            {rule.module_from for rule in self.real_rules.rules
            if rule.connection.name not in self.hidden_connections} | \
            {rule.module_to for rule in self.real_rules.rules
            if rule.connection.name not in self.hidden_connections}

    @property
    def modules_in_rules(self):
        return list(filter(lambda module: not self._module_not_in_rules(module), self.modules.values()))

    def _filter_files(self, files):
        return list(filter(lambda file: file.endswith('.py') and not file.startswith('__'), files))

    def _convert_path_to_module_path(self, path: str) -> str:
        return '.'.join(module.replace('.py', '') for module in path.split(os.sep)
                        if module not in self.root.split(os.sep))

    def _full_path_to_module(self, path: str, filename: str) -> str:
        return path + '/' + filename

    @property
    def code_root(self):
        if self._code_root == '':
            self.read_config()
        return self._code_root

    def _is_ignored(self, module_path: str):
        return any(map(lambda ignored: ignored in module_path, self.ignore))

    # TODO: Split into less complex chunks
    @property
    def modules(self):  # noqa: C901
        if self._modules:
            return self._modules
        for subdir, _, files in os.walk(self.root):
            files = self._filter_files(files)
            if len(files) == 0:
                continue
            for file in files:

                full_path = self._full_path_to_module(subdir, file)
                module_path = self._convert_path_to_module_path(full_path)
                if not module_path.startswith(self.code_root) or self._is_ignored(module_path):
                    continue

                if not self.is_package and self.code_root:
                    module_path = module_path.replace(self.code_root + '.', '')

                self._modules[module_path] = Module(full_path, module_path)

        return self._modules

    def _add_rules(self, module_from: Module, connection_paths: set[Path], connection_type: ConnectionType):
        for dependency in connection_paths:
            if dependency.module_path not in self.modules:
                continue
            connection = Connection(connection_type.value.uml,
                                    connection_type.value.name, dependency.lineno)
            self.real_rules.add(module_from, self.modules[dependency.module_path], connection)

    def read_project(self):
        self.read_config()

        for module in self.modules.values():
            self._add_rules(module, module.dependencies, ConnectionType.DEPENDENCY)
            self._add_rules(module, module.aggregations, ConnectionType.AGGREGATION)
            self._add_rules(module, module.extensions, ConnectionType.EXTENSION)

        self._filter_connection_priority()

    # TODO: Redo as set functionality
    def _filter_connection_priority(self):  # noqa: C901
        priority_line = ['aggregation', 'extension', 'dependency']
        connections = {key: set() for key in priority_line}
        for rule in self.real_rules.rules:
            connections[rule.connection.name].add(rule)

        for i in range(len(priority_line) - 1):
            current_connection = priority_line[i]  # aggregation
            lower_connections = priority_line[i + 1:]  # extension, dependency
            for rule in connections[current_connection]:
                for lower_connection in lower_connections:
                    for lower_rule in connections[lower_connection]:
                        if lower_rule not in self.real_rules.rules:
                            continue
                        if rule.module_from == lower_rule.module_from and rule.module_to == lower_rule.module_to:
                            self.real_rules.drop(lower_rule)

    def search_violations(self):
        for rule in self.real_rules.rules:
            self._check_rule_against_config(rule)

    def _check_rule_against_config(self, rule: Rule):
        if rule.module_from.path.path not in self.raw_rules.keys():
            self.violations.add(rule)
            return
        for module_path, constraints in self.raw_rules.items():
            if not self._is_inside_path(rule.module_from, module_path):
                continue
            self._check_rule_conformance_to_constraints(rule, constraints)

    def _check_rule_conformance_to_constraints(self, rule, constraints):
        if rule.connection.name not in constraints:
            self.violations.add(rule)
            return
        for constraint, target_modules in constraints.items():
            if not target_modules:
                self._handle_possible_violation(rule, constraint)
                continue
            self._search_non_conformant_rules(rule, target_modules, constraint)

    def _search_non_conformant_rules(self, rule, target_modules, constraint):
        for target_module in target_modules:
            if self._is_inside_path(rule.module_to, target_module) or \
                    self._is_module_specified_in_config(rule.module_to, target_modules):
                continue
            self._handle_possible_violation(rule, constraint)

    def _is_inside_path(self, module, path):
        return module.path.path.startswith(path)

    def _is_module_specified_in_config(self, module: Module, all_target_modules: list[str]):
        return any(map(lambda target: self._is_inside_path(module, target), all_target_modules))

    def _handle_possible_violation(self, rule, constraint):
        if rule.connection.name == constraint:
            self.violations.add(rule)



import sys

# Gracefuly taken from StackOverflow
# Colored printing functions for strings that use universal ANSI escape sequences.
# fail: bold red, pass: bold green, warn: bold yellow,
# info: bold blue, bold: bold white


class ColorPrint:

    @staticmethod
    def print_fail(message, end='\n'):
        sys.stderr.write('\x1b[1;31m' + message + '\x1b[0m' + end)

    @staticmethod
    def print_pass(message, end='\n'):
        sys.stdout.write('\x1b[1;32m' + message + '\x1b[0m' + end)

    @staticmethod
    def print_warn(message, end='\n'):
        sys.stderr.write('\x1b[1;33m' + message + '\x1b[0m' + end)

    @staticmethod
    def print_info(message, end='\n'):
        sys.stdout.write('\x1b[1;34m' + message + '\x1b[0m' + end)

    @staticmethod
    def print_header(message, end='\n'):
        sys.stdout.write('\033[95m' + message + '\033[0m' + end)

    @staticmethod
    def print_bold(message, end='\n'):
        sys.stdout.write('\x1b[1;37m' + message + '\x1b[0m' + end)



import os
import sys

from architest.model import Project
from architest.utils.colorprint import ColorPrint as color

VERSION_NUMBER = '0.1.7'


def print_fill(message: str, fill_char: str = '=', print_function=print):  # noqa: T002
    terminal_width = os.get_terminal_size().columns
    wing_width = round((terminal_width - len(message) - 4) / 2) - 1
    wing = fill_char * wing_width
    print_function(wing + f'  {message}  ' + wing + '\n')


def search_violations(project: Project):
    project.read_project()
    project.search_violations()
    return project.violations


def _normalized_module_name_for_puml(name: str):
    return name.replace('.', '__').replace(' ', '')


def generate_plant_uml(root: str, hidden_connections: list[str] = []):  # noqa: C901
    project = Project(root)
    project.hidden_connections = project.hidden_connections + hidden_connections
    project.read_project()

    print('@startuml')  # noqa: T001
    upper_modules = []
    for module in project.modules_in_rules:
        upper_module = module.path.upper_module_name
        if upper_modules and upper_module != upper_modules[-1]:
            print('\n\n}')  # noqa: T001
        if upper_module not in upper_modules:
            print(f'frame {upper_module}', end='')  # noqa: T001
            print('{\n\n')  # noqa: T001
            upper_modules.append(upper_module)

        print(f'rectangle {_normalized_module_name_for_puml(module.path.path)} [')  # noqa: T001
        print(f'  <b>{module.name}</b>')  # noqa: T001
        print('---')  # noqa: T001
        print(f'<i>{module.path.path}</i>')  # noqa: T001
        print('---')  # noqa: T001
        for doc in module.docstrings:
            print(f'"{doc}"')  # noqa: T001
            print()  # noqa: T001
        print(']')  # noqa: T001
        print()  # noqa: T001
    for rule in project.real_rules.rules:
        if rule.connection.name in project.hidden_connections:
            continue
        print(_normalized_module_name_for_puml(rule.module_from.path.path) + ' '  # noqa: T001
            + rule.connection.uml + ' '
            + _normalized_module_name_for_puml(rule.module_to.path.path))  # noqa: T001
    print('@enduml')  # noqa: T001


def main(root: str):
    print()  # noqa: T001
    print_fill('Starting Architest', print_function=color.print_bold)
    print_fill('firmitatis · utilitatis · venustatis', ' ')
    print(f'version: {VERSION_NUMBER} · target project: {root}\n')  # noqa: T001

    file_path = root + '/' + '.architest.yml'
    if not os.path.exists(file_path):
        print_fill('No .architest.yml file found at project root', print_function=color.print_warn)
        sys.exit(1)

    print_fill('Searching for design violations', '-')

    project = Project(root)
    violations = search_violations(project)
    if len(violations) > 0:
        for violation in violations:
            print(violation.violation_error())  # noqa: T001
        print()  # noqa: T001
        print_fill(f'Found {len(violations)} violation{"" if len(violations) == 1 else "s"}',
                   print_function=color.print_fail)
        sys.exit(1)
    else:
        print_fill('Design conforms to config', print_function=color.print_pass)

    print()  # noqa: T001
    sys.exit(0)


def entry_point(args: list[str]):
    if args[1] in {'-d', '--draw'}:
        additional_hidden_connections = []
        if len(args) > 3 and args[3] == '--hide-connections':
            additional_hidden_connections = args[4:]
        generate_plant_uml(args[2], additional_hidden_connections)
        sys.exit(0)
    main(args[1])



from setuptools import setup

VERSION_NUMBER = '0.1.7'

setup(
    name='architest',
    packages=['architest', 'architest.utils'],
    version=VERSION_NUMBER,
    license='MIT',
    description="Architest allows you to check your project's structure conformity to initial architecture",
    author='Vladimir Semenov',
    author_email='subatiq@gmail.com',
    url='https://github.com/VASemenov/architest',
    download_url='https://github.com/VASemenov/architest/archive/refs/tags/0.1.7.tar.gz',
    keywords=['architest', 'static', 'test', 'architecture', 'design'],
    install_requires=[
        'pyyaml',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
    ],
)



import ast
import itertools
import sys


def get_plain_imports(body):
    imports = [[(name.name, obj.lineno) for name in obj.names]
               for obj in body if isinstance(obj, ast.Import)]
    return list(itertools.chain.from_iterable(imports))


def get_imports_from(body):
    imports = [[(obj.module + '.' + name.name, obj.lineno) for name in obj.names]
               for obj in body if isinstance(obj, ast.ImportFrom)]
    return list(itertools.chain.from_iterable(imports))


def find_docstrings(filename):
    with open(filename) as f:
        data = f.read()
        module = ast.parse(data)
        body = module.body
        class_docs = [obj.name + ' class:\n' + ast.get_docstring(obj, clean=True) for obj in body if isinstance(
            obj, ast.ClassDef) and ast.get_docstring(obj, clean=True)]
        module_docstring = ast.get_docstring(module, clean=True)
        if module_docstring:
            module_docstring = 'module:\n' + module_docstring
        return list(filter(lambda doc: doc, [module_docstring] + class_docs))


def find_dependencies(filename: str):
    with open(filename) as f:
        data = f.read()
        module = ast.parse(data)
        return get_plain_imports(module.body) + get_imports_from(module.body)


def find_in_class_calls(filename: str):
    with open(filename) as f:
        data = f.read()
        module = ast.parse(data)
        classes = [
            obj
            for obj in module.body
            if isinstance(obj, ast.ClassDef)
        ]
        class_names = [obj.name for obj in classes]
        return [
            (node.func.id, node.func.lineno) for node in ast.walk(module)
            if (isinstance(node, ast.Call) and isinstance(node.func, ast.Name)
                and node.func.id not in class_names)
        ]


def find_instantiations(filename: str):
    with open(filename) as f:
        data = f.read()
        module = ast.parse(data)
        classes = [
            obj
            for obj in module.body
            if isinstance(obj, ast.ClassDef)
        ]
        class_names = [obj.name for obj in classes]
        return [
            (node.value.func.id, node.value.func.lineno) for node in ast.walk(module)
            if (isinstance(node, ast.Assign) and isinstance(node.value, ast.Call)
                and isinstance(node.value.func, ast.Name) and node.value.func.id[0].
                isupper() and node.value.func.id not in class_names)
        ]


def get_constants_in_base(base):
    if isinstance(base, ast.Subscript):
        if isinstance(base.slice, ast.Name):
            return {(base.slice.id, base.lineno)}
        return {(element.id, element.lineno) for element in base.slice.elts}
    elif isinstance(base, ast.Attribute):
        return {(base.value.id, base.lineno)}
    return {(base.id, base.lineno)}


def find_extensions(filename: str) -> list[str]:
    with open(filename) as f:
        data = f.read()
        module = ast.parse(data)
        classes = [
            obj
            for obj in module.body
            if isinstance(obj, ast.ClassDef)
        ]
        class_names = [obj.name for obj in classes]
        extensions = []
        for _class in classes:
            extensions += [get_constants_in_base(base) for base in _class.bases]

        return {(name, lineno)
                for name, lineno in itertools.chain.from_iterable(extensions)
                if name not in class_names}


if __name__ == '__main__':
    filename = sys.argv[1]
    find_dependencies(filename)
